import json
import time
import re
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.logger import get_ixnet_logger
from ixnetwork_restpy import BatchFind, BatchAdd


class Vport(object):
    """Transforms OpenAPI objects into IxNetwork objects

    Args
    ----
    - ixnetworkapi (Api): instance of the Api class

    Transformations
    ---------------
    - /components/schemas/Port to /vport
    - /components/schemas/Layer1 to /vport/l1Config/...

    Process
    -------
    - Remove any vports that are not in the config.ports
    - Add any vports that are in the config.ports
    - If the location of the config.ports.location is different than the
      the /vport -connectedTo property set it to None
    - If the config.ports.location is None don't connect the ports
      else connect the port, get the vport type, set the card mode based on the
      config.layer1.speed

    Notes
    -----
    - Uses resourcemanager to set the vport location and l1Config as it is the
      most efficient way. DO NOT use the AssignPorts API as it is too slow.
    - Only setup l1Config if location is connected.
    - Given a connected location and speed the vport -type, card resource mode
      and l1Config sub node are derived.

    """

    _SPEED_MAP = {
        "speed_800_gbps": "speed800g",
        "speed_400_gbps": "speed400g",
        "speed_200_gbps": "speed200g",
        "speed_100_gbps": "speed100g",
        "speed_50_gbps": "speed50g",
        "speed_40_gbps": "speed40g",
        "speed_25_gbps": "speed25g",
        "speed_10_gbps": "speed10g",
        "speed_1_gbps": "speed1000",
        "speed_100_fd_mbps": "speed100fd",
        "speed_100_hd_mbps": "speed100hd",
        "speed_10_fd_mbps": "speed10fd",
        "speed_10_hd_mbps": "speed10hd",
    }
    _VM_SPEED_MAP = {
        "speed_400_gbps": "speed400g",
        "speed_200_gbps": "speed200g",
        "speed_100_gbps": "speed100g",
        "speed_90_gbps": "speed90g",
        "speed_80_gbps": "speed80g",
        "speed_70_gbps": "speed70g",
        "speed_60_gbps": "speed60g",
        "speed_50_gbps": "speed50g",
        "speed_40_gbps": "speed40g",
        "speed_30_gbps": "speed30g",
        "speed_25_gbps": "speed25g",
        "speed_20_gbps": "speed20g",
        "speed_10_gbps": "speed10g",
        "speed_9_gbps": "speed9000",
        "speed_8_gbps": "speed8000",
        "speed_7_gbps": "speed7000",
        "speed_6_gbps": "speed6000",
        "speed_5_gbps": "speed5000",
        "speed_4_gbps": "speed4000",
        "speed_3_gbps": "speed3000",
        "speed_2_gbps": "speed2000",
        "speed_1_gbps": "speed1000",
        "speed_100_mbps": "speed100",
        "speed_100_fd_mbps": "speed100",
        "speed_100_hd_mbps": "speed100",
        "speed_10_fd_mbps": "speed100",
        "speed_10_hd_mbps": "speed100",
    }

    _SPEED_MODE_MAP = {
        "speed_1_gbps": "normal",
        "speed_10_gbps": "tengig",
        "speed_25_gbps": "twentyfivegig",
        "speed_40_gbps": "fortygig",
        "speed_50_gbps": "fiftygig",
        "speed_100_gbps": "^(?!.*(twohundredgig|fourhundredgig)).*hundredgig.*$",
        "speed_200_gbps": "twohundredgig",
        "speed_400_gbps": "fourhundredgig",
        "speed_800_gbps": "eighthundredgig",
    }

    _ADVERTISE_MAP = {
        "advertise_one_thousand_mbps": "speed1000",
        "advertise_one_hundred_fd_mbps": "speed100fd",
        "advertise_one_hundred_hd_mbps": "speed100hd",
        "advertise_ten_fd_mbps": "speed10fd",
        "advertise_ten_hd_mbps": "speed10hd",
    }
    _FLOW_CONTROL_MAP = {
        "ieee_802_1qbb": "ieee802.1Qbb",
        "ieee_802_3x": "ieee802.3x",
    }

    _RESULT_COLUMNS = [
        ("frames_tx", "Frames Tx.", int),
        ("frames_rx", "Valid Frames Rx.", int),
        ("frames_tx_rate", "Frames Tx. Rate", float),
        ("frames_rx_rate", "Valid Frames Rx. Rate", float),
        ("bytes_tx", "Bytes Tx.", int),
        ("bytes_rx", "Bytes Rx.", int),
        ("bytes_tx_rate", "Bytes Tx. Rate", float),
        ("bytes_rx_rate", "Bytes Rx. Rate", float),
        # ('pfc_class_0_frames_rx', 'Rx Pause Priority Group 0 Frames', int),
        # ('pfc_class_1_frames_rx', 'Rx Pause Priority Group 1 Frames', int),
        # ('pfc_class_2_frames_rx', 'Rx Pause Priority Group 2 Frames', int),
        # ('pfc_class_3_frames_rx', 'Rx Pause Priority Group 3 Frames', int),
        # ('pfc_class_4_frames_rx', 'Rx Pause Priority Group 4 Frames', int),
        # ('pfc_class_5_frames_rx', 'Rx Pause Priority Group 5 Frames', int),
        # ('pfc_class_6_frames_rx', 'Rx Pause Priority Group 6 Frames', int),
        # ('pfc_class_7_frames_rx', 'Rx Pause Priority Group 7 Frames', int),
    ]

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._layer1_check = []
        self._interval = 1
        self._timeout = 10
        self.logger = get_ixnet_logger(__name__)

    def config(self):
        """Transform config.ports into Ixnetwork.Vport
        1) delete any vport that is not part of the config
        2) create a vport for every config.ports[] not present in IxNetwork
        3) set config.ports[].location to /vport -location or -connectedTo
        4) set /vport/l1Config/... properties using the corrected /vport -type
        5) connectPorts to use new l1Config settings and clearownership
        """
        self.logger.debug("Configuring Vports")
        self._resource_manager = self._api._ixnetwork.ResourceManager
        self._ixn_vport = self._api._vport
        self._layer1_check = []
        self._api._ixnetwork.StopAllProtocols(arg1="sync")
        self._wait_for(
            lambda: self.is_protocols_stopped(),
            """"Protocols are not stopped in {} seconds""".format(
                self._interval * self._timeout
            ),
            self._interval,
            self._timeout,
        )
        with Timer(self._api, "Ports configuration"):
            self._delete_vports()
            self._create_vports()
        with Timer(self._api, "Captures configuration"):
            self._api.capture.config()
        with Timer(self._api, "Location configuration"):
            self._set_location()
        with Timer(self._api, "Layer1 configuration"):
            self._set_layer1()

    def _wait_for(self, func, exp_msg, interval, timeout):
        end_time = round(time.time()) + timeout
        while True:
            res = func()
            if round(time.time()) >= end_time:
                raise Exception(exp_msg)
            if res:
                return res
            time.sleep(interval)

    def is_protocols_stopped(self):
        topos = self._api._ixnetwork.Topology.find()
        stopped = True
        if len(topos) > 0:
            dgs = topos.DeviceGroup.find()
            if len(dgs) > 0:
                eth_list = dgs.Ethernet.find()
                if len(eth_list) > 0:
                    if len(eth_list.Ipv4.find()) > 0:
                        if any(
                            "up" in status or "down" in status
                            for status in dgs.Ethernet.find().SessionStatus
                        ):
                            stopped = False

        return stopped

    def set_link_state(self, link_state):
        self.logger.debug("Vport setting link state")
        with Timer(self._api, "Link State operation"):
            payload = {
                "arg1": [],
                "arg2": link_state.state,
            }
            for port_name in link_state.port_names:
                payload["arg1"].append(
                    self._api.ixn_objects.get_href(port_name)
                )
            url = "%s/vport/operations/linkupdn" % self._api._ixnetwork.href
            self._api._request("POST", url, payload)

    def _import(self, imports):
        self.logger.debug("Importing vport configs")
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False
            )
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True

    def _delete_vports(self):
        """Delete any vports from the api server that do not exist in the new config"""
        self.logger.debug("Deleting vports")
        self._api._remove(self._ixn_vport, self._api.snappi_config.ports)

    def _create_vports(self):
        """Add any vports to the api server that do not already exist"""
        self.logger.debug("Creating vports")
        vports = self._api.select_vports()
        imports = []
        for port in self._api.snappi_config.ports:
            if port.name not in vports.keys():
                index = len(vports) + len(imports) + 1
                vport_import = {
                    "xpath": "/vport[%i]" % index,
                    "name": port.name,
                    "rxMode": "captureAndMeasure",
                    "txMode": "interleaved",
                }
                location = port.get("location")
                if location is None:
                    vport_import["connectedTo"] = location
                    port.location = None
                imports.append(vport_import)
        self._import(imports)
        for name, vport in self._api.select_vports().items():
            self._api.ixn_objects.set(name, vport)

    def _add_hosts(self, HostReadyTimeout):
        self.logger.debug("Adding hosts in vport")
        chassis = self._api._ixnetwork.AvailableHardware.Chassis
        add_addresses, check_addresses = set(), set()
        existing_chassis = {c.Hostname for c in chassis.find()}
        ports = self._api.snappi_config.ports or []
        for port in ports:
            location = port.get("location")
            if location is not None:
                location_info = self._api.parse_location_info(location)
                chassis_address = location_info.chassis_info
                if chassis_address not in existing_chassis:add_addresses.add(chassis_address)
                check_addresses.add(chassis_address)
        chassis_chains = self._api.ixnet_specific_config.chassis_chains
        if chassis_chains:
            for chassis_chain in chassis_chains:
                add_addresses.add(chassis_chain.primary)
                check_addresses.add(chassis_chain.primary)
                for secondary in chassis_chain.secondary:
                    add_addresses.add(secondary.location)
                    check_addresses.add(secondary.location)
        if len(add_addresses) > 0:
            with Timer(
                self._api, "Add location hosts [%s]" % ", ".join(add_addresses)
            ):
                with BatchAdd(self._api._ixnetwork):
                    for add_address in add_addresses:
                        chassis.add(Hostname=add_address)
        if len(check_addresses) > 0:
            with Timer(
                self._api,
                "Location hosts ready [%s]" % ", ".join(check_addresses),
            ):
                start_time = time.time()
                while True:
                    chassis.find(
                        Hostname="^(%s)$" % "|".join(check_addresses),
                        State="^ready$",
                    )
                    if len(chassis) == len(check_addresses):
                        break
                    if time.time() - start_time > HostReadyTimeout:
                        raise RuntimeError(
                            "After %s seconds, not all location hosts [%s] are reachable"
                            % (HostReadyTimeout, ", ".join(check_addresses))
                        )
                    time.sleep(2)

    def _add_chassischain(self):
        chassis_chains = self._api.ixnet_specific_config.chassis_chains
        chassis = self._api._ixnetwork.AvailableHardware.Chassis

        self.logger.debug("Configuring Chassis Chain")
        if chassis_chains:
            for chassis_chain in chassis_chains:
                chassis.find(Hostname="^%s$" % chassis_chain.primary)
                if len(chassis) == 0:
                    self.logger.debug(
                        "Primary chassis [%s] not found"
                        % chassis_chain.primary
                    )
                    return
                chassis[0].ChainTopology = chassis_chain.topology
                for secondary in chassis_chain.secondary:
                    chassis.find(Hostname="^%s$" % secondary.location)
                    if len(chassis) == 0:
                        self.logger.debug(
                            "Secondary chassis [%s] not found"
                            % secondary.location
                        )
                        return
                    chassis[0].ChainTopology = chassis_chain.topology
                    chassis[0].PrimaryChassis = chassis_chain.primary
                    chassis[0].SequenceId = secondary.sequence_id
                    chassis[0].CableLength = secondary.cable_length
            self._api._ixnet_specific_config = None

    def _set_location(self):
        location_supported = True
        try:
            self._api._ixnetwork._connection._options(
                self._api._ixnetwork.href + "/locations"
            )
        except Exception:
            location_supported = False

        self._add_hosts(60)
        self._add_chassischain()
        with Timer(self._api, "Aggregation mode speed change"):
            layer1_check = self._api.resource_group.set_group()
            self._layer1_check.extend(layer1_check)
        vports = self._api.select_vports()
        locations = []
        imports = []
        clear_locations = []
        for port in self._api.snappi_config.ports:
            vport = vports[port.name]
            location = port.get("location")

            if location_supported is True:
                if vport["location"] == location and vport[
                    "connectionState"
                ].startswith("connectedLink"):
                    continue
            else:
                if len(vport["connectedTo"]) > 0 and vport[
                    "connectionState"
                ].startswith("connectedLink"):
                    continue

            # self._api.ixn_objects.set(port.name, vport)
            vport = {"xpath": vports[port.name]["xpath"]}
            if location_supported is True:
                vport["location"] = location
            else:
                if location is not None:
                    xpath = self._api.select_chassis_card_port(location)
                    vport["connectedTo"] = xpath
                else:
                    vport["connectedTo"] = ""
            imports.append(vport)
            if location is not None and len(location) > 0:
                clear_locations.append(location)
                locations.append(port.name)
        if len(locations) == 0:
            return
        self._clear_ownership(clear_locations)
        with Timer(self._api, "Location connect [%s]" % ", ".join(locations)):
            self._import(imports)
        with Timer(
            self._api, "Location state check [%s]" % ", ".join(locations)
        ):
            self._api._vport.find(ConnectionState="^(?!connectedLink).*$")
            if len(self._api._vport) > 0:
                self._api._vport.ConnectPorts()
            start = time.time()
            timeout = 10
            while True:
                self._api._vport.find(
                    Name="^(%s)$"
                    % "|".join(self._api.special_char(locations)),
                    ConnectionState="^connectedLink",
                )
                if len(self._api._vport) == len(locations):
                    break
                if time.time() - start > timeout:
                    unreachable = []
                    self._api._vport.find(
                        ConnectionState="^(?!connectedLink).*$"
                    )
                    for vport in self._api._vport:
                        unreachable.append(
                            "%s [%s: %s]"
                            % (
                                vport.Name,
                                vport.ConnectionState,
                                vport.ConnectionStatus,
                            )
                        )
                    raise RuntimeError(
                        "After %s seconds, %s are unreachable"
                        % (timeout, ", ".join(unreachable))
                    )
                time.sleep(2)
            for vport in self._api._vport.find(
                ConnectionState="^(?!connectedLinkUp).*$"
            ):
                self._api.warning(
                    "%s %s" % (vport.Name, vport.ConnectionState)
                )

    def _set_layer1(self):
        """Set the /vport/l1Config/... properties
        This should only happen if the vport connectionState is connectedLink...
        as it determines the ./l1Config child node.
        """
        layer1_config = self._api.snappi_config.get("layer1")
        if layer1_config is None:
            return
        if len(layer1_config) == 0:
            return
        reset_auto_negotiation = dict()
        # set and commit the card resource mode
        vports = self._api.select_vports()
        imports = []
        for layer1 in layer1_config:
            for port_name in layer1.port_names:
                self._set_card_resource_mode(
                    vports[port_name], layer1, imports
                )
        if self._import(imports) is False:
            # WARNING: this retry is because no reasonable answer as to why
            # changing card mode periodically fails with this opaque message
            # 'Releasing ownership on ports failed.'
            self._api.info("Retrying card resource mode change")
            self._import(imports)
        # set the vport type
        imports = []
        for layer1 in layer1_config:
            for port_name in layer1.port_names:
                self._set_vport_type(vports[port_name], layer1, imports)
        self._import(imports)
        vports = self._api.select_vports()
        # set the remainder of l1config properties
        imports = []
        for layer1 in layer1_config:
            for port_name in layer1.port_names:
                self._set_l1config_properties(
                    vports[port_name], layer1, imports
                )
        self._import(imports)
        # Due to dependency attribute (ieeeL1Defaults)
        # reset enableAutoNegotiation
        imports = []
        for layer1 in layer1_config:
            for port_name in layer1.port_names:
                vport = vports[port_name]
                if (
                    port_name in reset_auto_negotiation
                    and reset_auto_negotiation[port_name]
                ):
                    self._reset_auto_negotiation(vport, layer1, imports)
        self._import(imports)

    def _set_l1config_properties(self, vport, layer1, imports):
        """Set vport l1config properties"""
        if vport["connectionState"] not in [
            "connectedLinkUp",
            "connectedLinkDown",
        ]:
            return
        self._set_fcoe(vport, layer1, imports)

        self._set_auto_negotiation(vport, layer1, imports)

    def _set_card_resource_mode(self, vport, layer1, imports):
        """If the card has an aggregation mode set it according to the speed"""
        if (
            vport["connectionState"]
            not in ["connectedLinkUp", "connectedLinkDown"]
            or layer1.name in self._layer1_check
        ):
            return

        aggregation_mode = None
        if layer1.speed in Vport._SPEED_MODE_MAP:
            card = self._api.select_chassis_card(vport)
            mode = Vport._SPEED_MODE_MAP[layer1.speed]
            for available_mode in card["availableModes"]:
                if re.search(mode, available_mode.lower()) is not None:
                    aggregation_mode = available_mode
                    break
        if (
            aggregation_mode is not None
            and aggregation_mode != card["aggregationMode"]
        ):
            self._api.info(
                "Setting %s to resource mode %s"
                % (card["description"], aggregation_mode)
            )
            imports.append(
                {"xpath": card["xpath"], "aggregationMode": aggregation_mode}
            )

    def _set_auto_negotiation(self, vport, layer1, imports):
        if layer1.speed.endswith("_mbps") or layer1.speed == "speed_1_gbps":
            self._set_ethernet_auto_negotiation(vport, layer1, imports)
        else:
            self._set_gigabit_auto_negotiation(vport, layer1, imports)

    def _set_vport_type(self, vport, layer1, imports):
        """Set the /vport -type

        If flow_control is not None then the -type attribute should
        be switched to a type with the Fcoe extension if it is allowed.

        If flow_control is None then the -type attribute should
        be switched to a type without the Fcoe extension.
        """
        fcoe = False
        flow_control = layer1.get("flow_control")
        if flow_control is not None:
            fcoe = True
        vport_type = vport["type"]
        elegible_fcoe_vport_types = [
            "ethernet",
            "tenGigLan",
            "fortyGigLan",
            "tenGigWan",
            "hundredGigLan",
            "tenFortyHundredGigLan",
            "novusHundredGigLan",
            "novusTenGigLan",
            "krakenFourHundredGigLan",
            "aresOneFourHundredGigLan",
            "starFourHundredGigLan",
            "aresOneM",
        ]
        if fcoe is True and vport_type in elegible_fcoe_vport_types:
            vport_type = vport_type + "Fcoe"
        if fcoe is False and vport_type.endswith("Fcoe"):
            vport_type = vport_type.replace("Fcoe", "")
        if vport_type != vport["type"]:
            imports.append(
                {
                    "xpath": vport["xpath"] + "/l1Config",
                    "currentType": vport_type,
                }
            )
        return vport_type

    def _set_ethernet_auto_negotiation(self, vport, layer1, imports):
        advertise = []
        if layer1.speed == "speed_1_gbps":
            advertise.append(
                Vport._ADVERTISE_MAP["advertise_one_thousand_mbps"]
            )
        if layer1.speed == "speed_100_fd_mbps":
            advertise.append(
                Vport._ADVERTISE_MAP["advertise_one_hundred_fd_mbps"]
            )
        if layer1.speed == "speed_100_hd_mbps":
            advertise.append(
                Vport._ADVERTISE_MAP["advertise_one_hundred_hd_mbps"]
            )
        if layer1.speed == "speed_10_fd_mbps":
            advertise.append(Vport._ADVERTISE_MAP["advertise_ten_fd_mbps"])
        if layer1.speed == "speed_10_hd_mbps":
            advertise.append(Vport._ADVERTISE_MAP["advertise_ten_hd_mbps"])
        proposed_import = {
            "xpath": vport["xpath"]
            + "/l1Config/"
            + vport["type"].replace("Fcoe", ""),
            "speed": self._get_speed(vport, layer1),
            "media": layer1.get("media", with_default=True),
            "autoNegotiate": layer1.get("auto_negotiate", with_default=True),
            "speedAuto": advertise,
        }
        self._add_l1config_import(vport, proposed_import, imports)

    def _add_l1config_import(self, vport, proposed_import, imports):
        type = vport["type"].replace("Fcoe", "")
        l1config = vport["l1Config"][type]
        key_to_remove = []
        for key in proposed_import:
            if key == "xpath":
                continue
            if key not in l1config or l1config[key] == proposed_import[key]:
                key_to_remove.append(key)
        # add this constrain due to handle some specific use case (1G to 10G)
        if "speed" in key_to_remove and "speedAuto" not in key_to_remove:
            key_to_remove.remove("speed")
        for key in key_to_remove:
            proposed_import.pop(key)
        if len(proposed_import) > 0:
            imports.append(proposed_import)

    def _set_gigabit_auto_negotiation(self, vport, layer1, imports):
        advertise = []
        advertise.append(
            Vport._SPEED_MAP[layer1.get("speed", with_default=True)]
        )
        auto_field_name = "enableAutoNegotiation"
        if re.search("novustengiglan", vport["type"].lower()) is not None:
            auto_field_name = "autoNegotiate"
        # Due to ieeeL1Defaults dependency
        ieee_l1_defaults = layer1.get("ieee_media_defaults", with_default=True)
        if ieee_l1_defaults is None:
            ieee_l1_defaults = "True"
        ieee_media_defaults = {
            "xpath": vport["xpath"]
            + "/l1Config/"
            + vport["type"].replace("Fcoe", ""),
            "ieeeL1Defaults": ieee_l1_defaults,
        }
        self._add_l1config_import(vport, ieee_media_defaults, imports)
        auto_negotiation = layer1.get("auto_negotiation", with_default=True)
        rs_fec = auto_negotiation.get("rs_fec", with_default=True)
        link_training = auto_negotiation.get(
            "link_training", with_default=True
        )
        auto_negotiate = layer1.get("auto_negotiate", with_default=True)
        if auto_negotiate is None:
            auto_negotiate = "True"
        proposed_import = {
            "xpath": vport["xpath"]
            + "/l1Config/"
            + vport["type"].replace("Fcoe", ""),
            "speed": Vport._SPEED_MAP[layer1.speed],
            "{0}".format(auto_field_name): (
                False if auto_negotiate is None else auto_negotiate
            ),
            "enableRsFec": False if rs_fec is None else rs_fec,
            "linkTraining": False if link_training is None else link_training,
            "speedAuto": advertise,
        }
        proposed_import["media"] = layer1.get("media", with_default=True)
        self._add_l1config_import(vport, proposed_import, imports)

    def _get_speed(self, vport, layer1):
        if vport["type"] == "ethernetvm":
            return Vport._VM_SPEED_MAP[layer1.speed]
        else:
            return Vport._SPEED_MAP[layer1.speed]

    def _reset_auto_negotiation(self, vport, layer1, imports):
        if (
            layer1.speed.endswith("_mbps") is False
            and layer1.speed != "speed_1_gbps"
        ):
            imports.append(
                {
                    "xpath": vport["xpath"]
                    + "/l1Config/"
                    + vport["type"].replace("Fcoe", ""),
                    "enableAutoNegotiation": layer1.get(
                        "auto_negotiate", with_default=True
                    ),
                }
            )

    def _set_fcoe(self, vport, layer1, imports):
        flow_control = layer1.get("flow_control")
        if flow_control is None:
            return
        directed_address = flow_control.get(
            "directed_address", with_default=True
        )
        directed_address = "".join(directed_address.split(":"))
        l1_xpath = "%s/l1Config/%s" % (
            vport["xpath"],
            vport["type"].replace("Fcoe", ""),
        )
        imports.append(
            {"xpath": l1_xpath, "flowControlDirectedAddress": directed_address}
        )
        xpath = "%s/l1Config/%s/fcoe" % (
            vport["xpath"],
            vport["type"].replace("Fcoe", ""),
        )
        fcoe = {
            "xpath": xpath,
            "flowControlType": Vport._FLOW_CONTROL_MAP[flow_control.choice],
        }
        if flow_control.choice == "ieee_802_1qbb":
            pfc = flow_control.get("ieee_802_1qbb", with_default=True)
            pfc_delay = pfc.get("pfc_delay", with_default=True)
            fcoe["enablePFCPauseDelay"] = False if pfc_delay == 0 else True
            fcoe["pfcPauseDelay"] = pfc_delay
            fcoe["pfcPriorityGroups"] = [
                -1 if pfc.pfc_class_0 is None else pfc.pfc_class_0,
                -1 if pfc.pfc_class_1 is None else pfc.pfc_class_1,
                -1 if pfc.pfc_class_2 is None else pfc.pfc_class_2,
                -1 if pfc.pfc_class_3 is None else pfc.pfc_class_3,
                -1 if pfc.pfc_class_4 is None else pfc.pfc_class_4,
                -1 if pfc.pfc_class_5 is None else pfc.pfc_class_5,
                -1 if pfc.pfc_class_6 is None else pfc.pfc_class_6,
                -1 if pfc.pfc_class_7 is None else pfc.pfc_class_7,
            ]
            fcoe["priorityGroupSize"] = "priorityGroupSize-8"
            fcoe["supportDataCenterMode"] = True
        imports.append(fcoe)

    def _get_chassis_location_map(self):
        location_to_href_map = {}
        chassis = self._api._ixnetwork.AvailableHardware.Chassis.find()
        locations = self._api._ixnetwork.Locations.find()

        for ch in chassis:
            location_to_href_map[ch.Hostname] = ch.href

        for loc in locations:
            for p in loc.Ports.find():
                location_to_href_map[p.Location] = p.href

        return location_to_href_map

    def _clear_ownership(self, locations):
        try:
            force_ownership = (
                self._api.snappi_config.options.port_options.location_preemption
            )
        except Exception:
            force_ownership = False
        self.logger.debug("location_preemption is %s" % force_ownership)
        if force_ownership is True:
            self.logger.debug("We are clearing ownership")
            available_hardware_hrefs = {}
            location_hrefs = {}
            href_map = self._get_chassis_location_map()
            for location in locations:
                if ";" in location:
                    clp = location.split(";")
                    # chassis = (
                    #     self._api._ixnetwork.AvailableHardware.Chassis.find(
                    #         Hostname=clp[0]
                    #     )
                    # )
                    # if len(chassis) > 0:
                    #     available_hardware_hrefs[
                    #         location
                    #     ] = "%s/card/%s/port/%s" % (
                    #         chassis.href,
                    #         abs(int(clp[1])),
                    #         abs(int(clp[2])),
                    #     )
                    if clp[0] in href_map:
                        available_hardware_hrefs[location] = (
                            "%s/card/%s/port/%s"
                            % (
                                href_map.get(clp[0]),
                                abs(int(clp[1])),
                                abs(int(clp[2])),
                            )
                        )
                elif "/" in location:
                    appliance = location.split("/")[0]
                    # locations = self._api._ixnetwork.Locations
                    # locations.find(Hostname=appliance)
                    # if len(locations) == 0:
                    #     locations.add(Hostname=appliance)
                    # ports = locations.Ports.find(Location="^%s$" % location)
                    # if len(ports) > 0:
                    #     location_hrefs[location] = ports.href
                    if location in href_map:
                        location_hrefs[location] = href_map.get(location)
                    else:
                        self._api._ixnetwork.Locations.add(Hostname=appliance)
                        href_map = self._get_chassis_location_map()
                        location_hrefs[location] = href_map.get(location)
            self._api.clear_ownership(available_hardware_hrefs, location_hrefs)

    def _set_result_value(
        self, row, column_name, column_value, column_type=str
    ):
        if (
            len(self._column_names) > 0
            and column_name not in self._column_names
        ):
            return
        try:
            row[column_name] = column_type(column_value)
        except Exception:
            if column_type.__name__ in ["float", "int"]:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def results(self, request):
        """Return port results"""

        self._column_names = request.get("column_names")
        if self._column_names is None:
            self._column_names = []
        elif not isinstance(self._column_names, list):
            msg = "Invalid format of port_names passed {},\
                    expected list".format(
                self._column_names
            )
            raise Exception(msg)

        port_names = request.get("port_names")
        if port_names is None or len(port_names) == 0:
            port_names = [port.name for port in self._api._config.ports]
        elif not isinstance(port_names, list):
            msg = "Invalid format of port_names passed {},\
                    expected list".format(
                port_names
            )
            raise Exception(msg)

        self.logger.debug(
            "Extracting %s stats for these ports %s"
            % (self._column_names, port_names)
        )
        port_filter = {"property": "name", "regex": ".*"}
        port_filter["regex"] = "^(%s)$" % "|".join(
            self._api.special_char(port_names)
        )

        port_rows = dict()
        vports = self._api.select_vports(port_name_filters=[port_filter])
        for vport in vports.values():
            port_row = dict()
            self._set_result_value(port_row, "name", vport.get("name"))
            location = vport.get("location")
            if (
                vport.get("connectionState").startswith("connectedLink")
                is True
            ):
                location += ";connected"
            elif len(location) > 0:
                location += ";" + vport.get("connectionState")
            else:
                location = vport.get("connectionState")
            self._set_result_value(port_row, "location", location)
            self._set_result_value(
                port_row,
                "link",
                (
                    "up"
                    if vport["connectionState"] == "connectedLinkUp"
                    else "down"
                ),
            )
            self._set_result_value(port_row, "capture", "stopped")
            # init all columns with corresponding zero-values so that
            # the underlying dictionary contains all requested columns
            # in an event of unwanted exceptions
            for ext_name, _, typ in self._RESULT_COLUMNS:
                self._set_result_value(port_row, ext_name, 0, typ)

            port_rows[vport["name"]] = port_row

        try:
            table = self._api.assistant.StatViewAssistant("Port Statistics")
        except Exception:
            self._api.warning("Could not retrive the port statistics viewer")
            return list(port_rows.values())

        self.logger.debug("These are port results:")
        for row in table.Rows:
            vport_name = row["Port Name"]
            if vport_name is None:
                raise Exception("Could not retrive 'Port Name' from stats")
            port_row = port_rows.get(vport_name)
            self.logger.debug(str(port_row))
            if port_row is None:
                continue
            for ext_name, int_name, typ in self._RESULT_COLUMNS:
                try:
                    row_val = row[int_name]
                    self._set_result_value(port_row, ext_name, row_val, typ)
                except Exception:
                    # TODO print a warning maybe ?
                    pass
        return list(port_rows.values())



from snappi_ixnetwork.timer import Timer
import time


class ProtocolMetrics(object):
    # Currently paging of the statistic view is not handled
    # TODO Need to enhance when device groups statistics reach
    # more than one page.

    _SUPPORTED_PROTOCOLS_ = [
        "bgpv4",
        "bgpv6",
        "macsec",
        "mka",
        "rocev2_ipv4",
        "rocev2_ipv6",
        "isis",
    ]

    _TOPO_STATS = {
        "name": "name",
        "total": "sessions_total",
        "up": "sessions_up",
        "down": "sessions_down",
        "notStarted": "sessions_not_started",
    }

    _RESULT_COLUMNS = {
        "bgpv4": [
            ("name", "Device Group", str),
            ("session_state", "Status", str),
            ("session_flap_count", "Session Flap Count", int),
            ("routes_advertised", "Routes Advertised", int),
            ("routes_received", "Routes Rx", int),
            ("route_withdraws_sent", "Routes Withdrawn", int),
            ("route_withdraws_received", "Route Withdraws Rx", int),
            ("updates_sent", "Updates Tx", int),
            ("updates_received", "Updates Rx", int),
            ("opens_sent", "Opens Tx", int),
            ("opens_received", "Opens Rx", int),
            ("keepalives_sent", "KeepAlives Tx", int),
            ("keepalives_received", "KeepAlives Rx", int),
            ("notifications_sent", "Notifications Tx", int),
            ("notifications_received", "Notifications Rx", int),
        ],
        "bgpv6": [
            ("name", "Device Group", str),
            ("session_state", "Status", str),
            ("session_flap_count", "Session Flap Count", int),
            ("routes_advertised", "Routes Advertised", int),
            ("routes_received", "Routes Rx", int),
            ("route_withdraws_sent", "Routes Withdrawn", int),
            ("route_withdraws_received", "Route Withdraws Rx", int),
            ("updates_sent", "Updates Tx", int),
            ("updates_received", "Updates Rx", int),
            ("opens_sent", "Opens Tx", int),
            ("opens_received", "Opens Rx", int),
            ("keepalives_sent", "KeepAlives Tx", int),
            ("keepalives_received", "KeepAlives Rx", int),
            ("notifications_sent", "Notifications Tx", int),
            ("notifications_received", "Notifications Rx", int),
        ],
        "macsec": [
            ("name", "Device Group", str),
            ("session_state", "Status", str),
            ("session_flap_count", "Session Flap Count", int),
            ("out_pkts_protected", "Protected Packet Tx", int),
            ("out_pkts_encrypted", "Encrypted Packet Tx", int),
            ("in_pkts_ok", "Valid Packet Rx", int),
            ("in_pkts_bad", "Total number of bad Packet Rx", int),
            ("in_pkts_bad_tag", "Bad Tag/ICV Discarded", int),
            ("in_pkts_late", "Out of Window Discarded", int),
            ("in_pkts_no_sci", "Unknown SCI Discarded", int),
            ("in_pkts_not_using_sa", "Unused SA Discarded", int),
            ("in_pkts_not_valid", "Invalid ICV Discarded", int),
            ("in_pkts_unknown_sci", "Unknown SCI Rx", int),
            ("in_pkts_unused_sa", "Unused SA Rx", int),
            ("in_pkts_invalid", "Invalid ICV Rx", int),
            ("in_pkts_untagged", "Non-MACsec Packet Rx", int),
            ("out_octets_protected", "Tx Bytes Protected", int),
            ("out_octets_encrypted", "Tx Bytes Encrypted", int),
            ("in_octets_validated", "Rx Bytes Validated", int),
            ("in_octets_decrypted", "Rx Bytes Decrypted", int),
        ],
        "mka": [
            ("name", "Device Group", str),
            ("session_state", "Status", str),
            ("mkpdu_tx", "MKPDU Tx", int),
            ("mkpdu_rx", "MKPDU Rx", int),
            ("live_peer_count", "Live Peer Count", int),
            ("potential_peer_count", "Potential Peer Count", int),
            ("latest_key_tx_peer_count", "Latest Key Tx Peer Count", int),
            ("latest_key_rx_peer_count", "Latest Key Rx Peer Count", int),
            ("malformed_mkpdu", "Malformed MKPDU", int),
            ("icv_mismatch", "ICV Mismatch", int),
        ],
        "rocev2_ipv4": [
            ("qp_configured", "QP Configured", int),
            ("qp_up", "QP Up", int),
            ("qp_down", "QP Down", int),
            ("connect_request_tx", "Connect Request Tx", int),
            ("connect_request_rx", "Connect Request Rx", int),
            ("connect_reply_tx", "Connect Reply Tx", int),
            ("connect_reply_rx", "Connect Reply Rx", int),
            ("ready_tx", "Ready Tx", int),
            ("ready_rx", "Ready Rx", int),
            ("disconnect_request_tx", "Disconnect Request Tx", int),
            ("disconnect_request_rx", "Disconnect Request Rx", int),
            ("disconnect_reply_tx", "Disconnect Reply Tx", int),
            ("disconnect_reply_rx", "Disconnect Reply Rx", int),
            ("reject_tx", "Reject Rx", int),
            ("reject_rx", "Reject Tx", int),
            ("unknown_msg_rx", "Unknown MSG Rx", int),
        ],
        "rocev2_ipv6": [
            ("qp_configured", "QP Configured", int),
            ("qp_up", "QP Up", int),
            ("qp_down", "QP Down", int),
            ("connect_request_tx", "Connect Request Tx", int),
            ("connect_request_rx", "Connect Request Rx", int),
            ("connect_reply_tx", "Connect Reply Tx", int),
            ("connect_reply_rx", "Connect Reply Rx", int),
            ("ready_tx", "Ready Tx", int),
            ("ready_rx", "Ready Rx", int),
            ("disconnect_request_tx", "Disconnect Request Tx", int),
            ("disconnect_request_rx", "Disconnect Request Rx", int),
            ("disconnect_reply_tx", "Disconnect Reply Tx", int),
            ("disconnect_reply_rx", "Disconnect Reply Rx", int),
            ("reject_tx", "Reject Rx", int),
            ("reject_rx", "Reject Tx", int),
            ("unknown_msg_rx", "Unknown MSG Rx", int),
        ],
        "isis": [
            ("name", "Device Group", str),
            ("l1_sessions_up", "L1 Sessions Up", int),
            ("l1_session_flap", "L1 Session Flap", int),
            ("l1_database_size", "L1 DB Size", int),
            ("l1_broadcast_hellos_sent", "L1 Hellos Tx", int),
            ("l1_broadcast_hellos_received", "L1 Hellos Rx", int),
            ("l1_point_to_point_hellos_sent", "L1 P2P Hellos Tx", int),
            ("l1_point_to_point_hellos_received", "L1 P2P Hellos Rx", int),
            ("l1_psnp_sent", "L1 PSNP Tx", int),
            ("l1_psnp_received", "L1 PSNP Rx", int),
            ("l1_csnp_sent", "L1 CSNP Tx", int),
            ("l1_csnp_received", "L1 CSNP Rx", int),
            ("l1_lsp_sent", "L1 LSP Tx", int),
            ("l1_lsp_received", "L1 LSP Rx", int),
            ("l2_sessions_up", "L2 Sessions Up", int),
            ("l2_session_flap", "L2 Session Flap", int),
            ("l2_database_size", "L2 DB Size", int),
            ("l2_broadcast_hellos_sent", "L2 Hellos Tx", int),
            ("l2_broadcast_hellos_received", "L2Hellos Rx", int),
            ("l2_point_to_point_hellos_sent", "L2 P2P Hellos Tx", int),
            ("l2_point_to_point_hellos_received", "L2 P2P Hellos Rx", int),
            ("l2_psnp_sent", "L2 PSNP Tx", int),
            ("l2_psnp_received", "L2 PSNP Rx", int),
            ("l2_csnp_sent", "L2 CSNP Tx", int),
            ("l2_csnp_received", "L2 CSNP Rx", int),
            ("l2_lsp_sent", "L2 LSP Tx", int),
            ("l2_lsp_received", "L2 LSP Rx", int),
        ],
    }

    _PROTO_NAME_MAP_ = {
        "bgpv4": {
            "per_port": "BGP Peer Per Port",
            "drill_down": "BGP Peer Drill Down",
            "drill_down_options": [
                "BGP Peer:Per Device Group",
                "BGP Peer:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["bgpv4"]],
            "ixn_name": "bgpIpv4Peer",
        },
        "bgpv6": {
            "per_port": "BGP\+ Peer Per Port",
            "drill_down": "BGP\+ Peer Drill Down",
            "drill_down_options": [
                "BGP+ Peer:Per Device Group",
                "BGP+ Peer:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["bgpv6"]],
            "ixn_name": "bgpIpv6Peer",
        },
        "macsec": {
            "per_port": "Static MACsec Per Port",
            "drill_down": "Static MACsec Drill Down",
            "drill_down_options": [
                "Static MACsec:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["macsec"]],
            "ixn_name": "staticMacsec",
        },
        "mka": {
            "per_port": "MKA Per Port",
            "drill_down": "MKA Drill Down",
            "drill_down_options": [
                "MKA:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["mka"]],
            "ixn_name": "Mka",
        },
        "rocev2_ipv4": {
            "per_port": "RoCEv2 Per Port",
            "drill_down": "RoCEv2 Drill Down",
            "drill_down_options": [
                "RoCEv2:Per Device Group",
                "RoCEv2:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["rocev2_ipv4"]],
            "ixn_name": "rocev2",
        },
        "rocev2_ipv6": {
            "per_port": "RoCE6v2 Per Port",
            "drill_down": "RoCE6v2 Drill Down",
            "drill_down_options": [
                "RoCE6v2:Per Device Group",
                "RoCE6v2:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["rocev2_ipv6"]],
            "ixn_name": "roce6v2",
        },
        "isis": {
            "per_port": "ISIS-L3 RTR Per Port",
            "drill_down": "ISIS-L3 RTR Drill Down",
            "drill_down_options": [
                "ISIS-L3 RTR:Per Device Group",
                "ISIS-L3 RTR:Per Session",
            ],
            "supported_stats": [s[0] for s in _RESULT_COLUMNS["isis"]],
            "ixn_name": "isisL3",
        },
    }

    _PEER_NAMES = {
        "bgpv4": "peer_names",
        "bgpv6": "peer_names",
        "macsec": "secure_entity_names",
        "mka": "peer_names",
        "rocev2_ipv4": "peer_names",
        "rocev2_ipv6": "peer_names",
        "isis": "router_names"
    }

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self.ixn = None
        self.columns = []
        self.device_names = []
        self.metric_timeout = 90
        self.interval = 1

    def _get_search_payload(self, parent, child, properties, filters):
        payload = {
            "selects": [
                {
                    "from": parent,
                    "properties": [],
                    "children": [
                        {
                            "child": child,
                            "properties": properties,
                            "filters": filters,
                        }
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "{}/operations/select?xpath=true".format(self.ixn.href)
        return (url, payload)

    def _select_view(self, view_filters=[]):
        url, payload = self._get_search_payload(
            "/statistics", "view", ["caption"], view_filters
        )
        result = self.ixn._connection._execute(url, payload)[0]
        return result.get("view")

    def _wait_for(self, func, exp_msg, interval, timeout):
        end_time = round(time.time()) + timeout
        while True:
            res = func()
            if round(time.time()) >= end_time:
                raise Exception(exp_msg)
            if res is not None:
                return res
            time.sleep(interval)

    def get_supported_protocols(self):
        """
        Return the protocols that are supported currently
        """
        return self._SUPPORTED_PROTOCOLS_

    def _port_list_in_per_port(self, protocol):
        self.ixn = self._api.assistant._ixnetwork
        protocol_name = self._PROTO_NAME_MAP_.get(protocol)
        if protocol_name is None:
            raise NotImplementedError("{} is Not Implemented".format(protocol))
        filter = [
            {
                "property": "caption",
                "regex": "^%s$" % protocol_name["per_port"],
            }
        ]
        view = self._wait_for(
            lambda: self._select_view(filter),
            "could not retrieve the view for %s" % protocol,
            self.interval,
            self.metric_timeout,
        )
        res = self._get_column_values(view[0]["href"], "Port")
        if res is None or len(res) == 0:
            raise Exception("Could not retrieve the Port column")
        return (res, view[0])

    def _get_column_values(self, href, value):
        payload = {"arg1": href, "arg2": value}
        url = "{}/statistics/view/operations/getcolumnvalues".format(
            self.ixn.href
        )
        res = self._api._request("POST", url, payload)
        return res.get("result")

    def _get_value(self, href, row_label, column_label):
        payload = {"arg1": href, "arg2": row_label, "arg3": column_label}
        url = "{}/statistics/view/operations/getvalue".format(self.ixn.href)
        res = self._api._request("POST", url, payload)
        return res.get("result")

    def _check_if_page_ready(self, view):
        count = 0
        while True:
            view.Refresh()
            if view.Data.IsReady:
                break
            if count >= self.metric_timeout:
                raise Exception("View Page is not ready")
            time.sleep(0.5)
            count += 1

    def _port_names_from_devices(self):
        config = self._api.snappi_config
        if len(self.device_names) == 0:
            port_list = [p.name for p in config.ports]
            lag_list = [lag.name for lag in config.lags]
            port_list = port_list + lag_list
            return port_list
        port_list = []
        for dev in config.devices:
            ethernets = dev.get("ethernets")
            if ethernets is None:
                continue
            for eth in ethernets:
                if eth.get("connection") is not None:
                    port_list.append(eth.get("connection").get("port_name"))
        return port_list

    def _do_drill_down(self, view, per_port, row_index, drill_option):
        url, payload = self._get_search_payload(
            view["href"],
            "drillDown",
            ["targetDrillDownOption", "targetRowIndex"],
            [],
        )
        result = self.ixn._connection._execute(url, payload)[0]
        if result.get("drillDown") is None:
            raise Exception("Could not fetch drill down node")

        payload = {
            "targetDrillDownOption": drill_option,
            "targetRowIndex": row_index,
        }
        url = result["drillDown"]["href"]
        count = 0
        while count < 5:
            # retrying as the linux api server throws error for first 2
            # consecutive executions
            try:
                self._api._request("PATCH", url, payload)
                break
            except Exception:
                self._check_if_page_ready(
                    self.ixn.Statistics.View.find(Caption=per_port)
                )
                time.sleep(0.3)
                count += 1

        url = "{}/statistics/view/drillDown/operations/dodrilldown".format(
            self.ixn.href
        )
        payload = {"arg1": result["drillDown"]["href"]}
        self._api._request("POST", url, payload)
        return

    def _get_per_device_group_stats(self, protocol):
        ports, v = self._port_list_in_per_port(protocol)
        config_ports = self._port_names_from_devices()
        indices = set(
            [ports.index(p) for p in list(set(config_ports)) if p in ports]
        )
        drill_options = self._PROTO_NAME_MAP_[protocol].get(
            "drill_down_options", []
        )
        drill_name = self._PROTO_NAME_MAP_[protocol]["drill_down"]
        per_port = self._PROTO_NAME_MAP_[protocol]["per_port"]
        column_names = self._RESULT_COLUMNS.get(protocol, [])
        row_lst = list()
        for i in indices:
            try:
                drill = self.ixn.Statistics.View.find(Caption=drill_name)
                for option in drill_options:
                    self._do_drill_down(v, per_port, i, option)
                self._check_if_page_ready(drill)
            except Exception as e:
                msg = """
                Could not retrive drill down view \
                at row index {} {}""".format(
                    i, e
                )
                raise Exception(msg)
            columns = drill.Data.ColumnCaptions
            drill.Data.PageSize = drill.Data.TotalRows
            values = drill.Data.PageValues
            for value in values:
                row_dt = dict()
                data = dict(zip(columns, value[0]))
                for col in column_names:
                    sn, ixn, typ = col[:3]
                    skip = False if len(col) <= 3 else col[-1]
                    self._set_result_value(row_dt, data, sn, ixn, typ, skip)
                if row_dt != {}:
                    row_lst.append(row_dt)
        return row_lst

    def _update_actual_dev_name(self, data):
        keys = self._api.dev_compacted.keys()
        if data["Device Group"] in keys:
            for k, v in self._api.dev_compacted.items():
                if (
                    data["Device Group"] == v["dev_name"]
                    and int(data["Device#"]) == v["index"] + 1
                ):
                    data["Device Group"] = k
        return data

    def _set_result_value(
        self, row_dt, data, stat_name, ix_name, stat_type=str, skip=False
    ):
        data = self._update_actual_dev_name(data)
        if self.device_names == []:
            self.device_names = [
                d.name for d in self._api.snappi_config.devices
            ]
        if data["Device Group"] not in self.device_names:
            return
        if skip:
            warn = stat_name in self.columns
            (
                self._api.warning(
                    "{} metric has no implementation".format(stat_name)
                )
                if warn
                else None
            )
            row_dt[stat_name] = (
                0 if stat_type.__name__ in ["float", "int"] else "na"
            )
        if len(self.columns) == 0 or stat_name in self.columns:
            try:
                row_dt[stat_name] = stat_type(data[ix_name])
                if isinstance(row_dt[stat_name], str):
                    row_dt[stat_name] = row_dt[stat_name].lower()
            except Exception:
                row_dt[stat_name] = (
                    0
                    if stat_type.__name__ in ["float", "int"]
                    else data[ix_name]
                )

    def _topo_stats(self, protocol):
        url = "%s/operations/gettopologystatus" % self.ixn.href
        res = self._api._request("POST", url)
        stats_map = {}
        for d in res["result"]:
            if self._PROTO_NAME_MAP_[protocol]["ixn_name"] not in d["arg1"]:
                continue
            stats_map[d["arg1"]] = {
                self._TOPO_STATS[i["arg1"]]: i["arg2"] for i in d["arg2"]
            }
        myfilter = [
            {
                "property": "name",
                "regex": (
                    ".*"
                    if len(self.device_names) == 0
                    else "^%s$"
                    % "|".join(self._api.special_char(self.device_names))
                ),
            }
        ]
        url, payload = self._get_search_payload(
            "/topology", "(?i)^(deviceGroup)$", ["name"], myfilter
        )
        result = self.ixn._connection._execute(url, payload)
        for t in result:
            if t.get("deviceGroup") is None:
                continue
            for d in t["deviceGroup"]:
                for p in stats_map:
                    if d["href"] not in p:
                        continue
                    stats_map[p]["name"] = d["name"]
        if self.device_names == []:
            return [stats_map[p] for p in stats_map]
        return [
            stats_map[p]
            for p in stats_map
            if stats_map["name"] in self.device_names
        ]

    def _filter_stats(self, protocol):
        self.ixn = self._api.assistant._ixnetwork
        f = [
            i
            for i in filter(
                lambda y: y not in self._TOPO_STATS.values(), self.columns
            )
        ]
        if len(f) > 0:
            return self._get_per_device_group_stats(protocol)
        else:
            return self._topo_stats(protocol)

    def results(self, request):
        """
        Return the Protocol statistics
        args:
            request: <type obj> MetricRequest object
            protocol: <type str> protocol name
        *Note: supported protocols can be fetched via,
        self.get_supported_protocols()
        """
        protocol = request.choice
        request = getattr(request, protocol)
        peer_name = self._PEER_NAMES[protocol]
        request.get(peer_name)
        self.device_names = request.get(peer_name)
        if self.device_names is None:
            self.device_names = []
        self.columns = request.get("column_names")
        if self.columns is None or self.columns == []:
            self.columns = self._PROTO_NAME_MAP_[protocol]["supported_stats"]
        if len(self.columns) > 0:
            self.columns.append("name") if "name" not in self.columns else None
            self.columns = list(set(self.columns))
        # with Timer(self._api, "Fetching {} Metrics".format(protocol)):
        return self._filter_stats(protocol)



import sys
import logging

APP_LOGGER_NAME = "snappi_ixnetwork"


def setup_ixnet_logger(log_level, file_name=None, module_name=None):
    logger = logging.getLogger(APP_LOGGER_NAME)
    logger.setLevel(log_level)
    formatter = logging.Formatter(
        fmt="%(asctime)s [%(name)s] [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    sh = logging.StreamHandler(sys.stdout)
    sh.setFormatter(formatter)
    if len(logger.handlers) > 0:
        del logger.handlers[:]
    logger.addHandler(sh)
    if file_name:
        fh = logging.FileHandler(file_name)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    if module_name is not None:
        logger = get_ixnet_logger(module_name)
    return logger


def get_ixnet_logger(module_name):
    module_name = ".".join(str(module_name).split(".")[1:])
    return logging.getLogger(APP_LOGGER_NAME).getChild(module_name)



import pytest


@pytest.mark.skip("skip until migrated to snappi")
def test_json_config(serializer, api, tx_port, rx_port, b2b_devices):
    config = Config(ports=[tx_port, rx_port], devices=b2b_devices)
    state = State(ConfigState(config=config, state="set"))
    state = serializer.json(state)
    api.set_state(state)


@pytest.mark.skip("skip until migrated to snappi")
def test_dict_config(serializer, api, tx_port, rx_port, b2b_devices):
    config = Config(ports=[tx_port, rx_port], devices=b2b_devices)
    state = State(ConfigState(config=config, state="set"))
    state = serializer.json_to_dict(serializer.json(state))
    api.set_state(state)


@pytest.mark.skip("skip until migrated to snappi")
def test_config(serializer, api, tx_port, rx_port, b2b_devices):
    config = Config(ports=[tx_port, rx_port], devices=b2b_devices)
    state = State(ConfigState(config=config, state="set"))
    api.set_state(state)


if __name__ == "__main__":
    pytest.main(["-s", __file__])



# class Events(object):
#     def __init__(self, ixnetworkapi):
#         self._api = ixnetworkapi
#
#     def config(self):
#         events = self._api.snappi_config._properties.get("events")
#         if events is None:
#             return
#         ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
#         if self.event_allow("rx_rate_threshold"):
#             if self._api.traffic_item.has_latency is True:
#                 raise Exception(
#                     "We are supporting either latency or rx_rate_threshold"
#                 )
#             ixn_CpdpConvergence.Enabled = True
#             ixn_CpdpConvergence.EnableControlPlaneEvents = True
#             ixn_CpdpConvergence.EnableDataPlaneEventsRateMonitor = True
#             if events.rx_rate_threshold.threshold is not None:
#                 ixn_CpdpConvergence.DataPlaneThreshold = (
#                     events.rx_rate_threshold.threshold
#                 )
#         else:
#             ixn_CpdpConvergence.Enabled = False
#
#     def event_allow(self, event_name):
#         events = self._api.snappi_config._properties.get("events")
#         if events is None:
#             return False
#         event = events._properties.get(event_name)
#         if events.enable is True or (
#             event is not None and event.enable is True
#         ):
#             return True
#         else:
#             return False



import json
import time
import re
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.logger import get_ixnet_logger
import snappi


class IxNetworkConfig(object):
    """
    Represents the IxNetwork configuration, managing chassis chains.
    ixnconfig = api.ixnet_specific_config
    chassis_chain1 = ixnconfig.chassis_chains.add()
    chassis_chain1.primary = "10.36.78.236"
    chassis_chain1.topology = chassis_chain1.STAR
    secondary1 = chassis_chain1.secondary.add()
    secondary1.location = "10.36.78.141"
    secondary1.sequence_id = "2"
    secondary1.cable_length = "6"
    """

    _TYPE = {"chassis_chains": {"type": "ChassisChainIter"}}

    def __init__(self, ixnetworkapi):
        """
        Initializes the IxNetworkConfig object.

        :param ixnetworkapi: API object for interacting with IxNetwork.
        :param chassis_chains: Object to add new chassis chain objects
        """
        print(f"End {object.__class__}")
        self._api = ixnetworkapi
        self._chassis_chains = None

    @property
    def chassis_chains(self):
        """
        Retrieves the chassis chains configuration.

        :return: ChassisChainIter object.
        """
        if self._chassis_chains is None:
            self._chassis_chains = ChassisChainIter()

        return self._chassis_chains

    def config(self):
        """
        Logs configuration settings for IxNetwork.
        """
        self.logger.info("Configuring IxNConfig")


class ChassisChain(object):
    """
    Represents a chassis chain configuration in the network.
    """

    _TYPES = {
        "primary": {"type": str},
        "topology": {"type": str},
        "enum": [
            "daisy",
            "star",
        ],
        "secondary": {"type": "SecondaryIter"},
    }

    _REQUIRED = ("primary",)

    DAISY = "daisy"
    STAR = "star"

    def __init__(self, primary=None, topology=None):
        """
        Initializes a ChassisChain object.

        :param primary: Primary chassis.
        :param topology: Network topology (daisy or star).
        """
        super(ChassisChain, self).__init__()
        self._primary = primary
        self._topology = topology
        self._secondary = None

    def set(self, primary=None, topology=None):
        """
        Sets the primary chassis and topology.

        :param primary: Primary chassis.
        :param topology: Network topology (daisy or star).
        """
        self._topology = topology
        self._primary = primary

    @property
    def primary(self):
        """
        Gets the primary chassis.
        """
        return self._primary

    @primary.setter
    def primary(self, value):
        """
        Sets the primary chassis.

        :param value: Primary chassis value.
        :raises TypeError: If value is None.
        """
        if value is None:
            raise TypeError("Cannot set required property primary as None")
        self._primary = value

    @property
    def topology(self):
        """
        Gets the topology type.
        """
        return self._topology

    @topology.setter
    def topology(self, value):
        """
        Sets the topology type.

        :param value: Topology type (daisy or star).
        :raises TypeError: If value is None.
        """
        if value is None:
            raise TypeError("Cannot set required property topology as None")
        self._topology = value

    @property
    def secondary(self):
        """
        Retrieves the secondary chassis configuration.
        """
        if self._secondary is None:
            self._secondary = SecondaryIter()

        return self._secondary


class ChassisChainIter(snappi.snappi.OpenApiIter):
    """
    Iterator class for managing multiple ChassisChain objects.
    """

    def __init__(self):
        super(ChassisChainIter, self).__init__()

    def __getitem__(self, key):
        return self._getitem(key)

    def __iter__(self):
        return self._iter()

    def __next__(self):
        return self._next()

    def next(self):
        return self._next()

    def _instanceOf(self, item):
        """
        Ensures that the item is an instance of ChassisChain.
        """
        if not isinstance(item, ChassisChain):
            raise Exception("Item is not an instance of ChassisChain")

    def chassisChain(self, primary=None, topology=None):
        """
        Adds a new ChassisChain object to the iterator.
        """
        item = ChassisChain(primary=primary, topology=topology)
        self._add(item)
        return self

    def add(self, primary=None, topology=None):
        """
        Adds a new ChassisChain instance to the list.
        """
        item = ChassisChain(primary=primary)
        self._add(item)
        return item


class Secondary(object):
    """
    Initializes a Secondary chassis object.

    :param location: Secondary chassis.
    :param sequence_id: Sequence ID of the secondary chassis.
    :param cable_length: Cable Length of the secondary chassis.
    """

    """
    Represents a secondary chassis configuration.
    """
    _TYPES = {
        "location": {"type": str},
        "sequence_id": {"type": int},
        "cable_length": {"type": int},
    }

    _REQUIRED = ("location",)

    def __init__(self, location=None, sequence_id=None, cable_length=None):
        super(Secondary, self).__init__()
        self._location = location
        self._sequence_id = sequence_id
        self._cable_length = cable_length

    def set(self, location=None, sequence_id=None, cable_length=None):
        self._location = location
        self._sequence_id = sequence_id
        self._cable_length = cable_length

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value is None:
            raise TypeError("Cannot set required property location as None")
        self._location = value

    @property
    def sequence_id(self):
        return self._sequence_id

    @sequence_id.setter
    def sequence_id(self, value):
        if value is None:
            raise TypeError("Cannot set required property sequence_id as None")
        self._sequence_id = value

    @property
    def cable_length(self):
        return self._cable_length

    @cable_length.setter
    def cable_length(self, value):
        if value is None:
            raise TypeError(
                "Cannot set required property cable_length as None"
            )
        self._cable_length = value


class SecondaryIter(snappi.snappi.OpenApiIter):
    """
    Iterator class for managing multiple Secondary objects.
    """

    def __init__(self):
        super(SecondaryIter, self).__init__()

    def __getitem__(self, key):
        return self._getitem(key)

    def __iter__(self):
        return self._iter()

    def __next__(self):
        return self._next()

    def next(self):
        return self._next()

    def _instanceOf(self, item):
        if not isinstance(item, Secondary):
            raise Exception("Item is not an instance of Port")

    def secondary(self, location=None, sequence_id=None, cable_length=None):
        item = Secondary(
            location=location,
            sequence_id=sequence_id,
            cable_length=cable_length,
        )
        self._add(item)
        return self

    def add(self, location=None, sequence_id=None, cable_length=None):
        item = Secondary(
            location=location,
            sequence_id=sequence_id,
            cable_length=cable_length,
        )
        self._add(item)
        return item



from snappi_ixnetwork.logger import get_ixnet_logger


class IxNetObjects(object):
    def __init__(self, ixnetworkapi):
        self._ixnet_infos = {}
        self.logger = get_ixnet_logger(__name__)
        self._api = ixnetworkapi

    # get_ixn_href
    def get_href(self, name):
        """Returns an href given a unique configuration name"""
        obj = self.get(name)
        self.logger.debug("get_href %s : %s" % (name, obj.href))
        return obj.href

    def get_xpath(self, name):
        """Returns an xpath given a unique configuration name"""
        obj = self.get(name)
        self.logger.debug("get_xpath %s : %s" % (name, obj.xpath))
        return obj.xpath

    def get_object(self, name):
        """Returns an internal ixnobject given a unique configuration name"""
        obj = self.get(name)
        return obj.ixnobject

    @property
    def names(self):
        """Returns all names stored as keys"""
        return self._ixnet_infos.keys()

    def get(self, name):
        try:
            return self._ixnet_infos[name]
        except KeyError:
            self.logger.debug("These are existing names %s" % self.names)
            raise NameError(
                "snappi object named {0} not found in internal db".format(name)
            )

    def get_working_dg(self, name):
        ixn_obj = self.get(name)
        return ixn_obj.working_dg

    def set(self, name, ixnobject):
        self._ixnet_infos[name] = IxNetInfo(
            ixnobject, self._api.ngpf.working_dg
        )

    def set_scalable(self, ixnobject):
        names = ixnobject.get("name")
        self.logger.debug("set_scalable names : %s" % names)
        set_names = []
        for index, name in enumerate(names):
            if name is None or name in set_names:
                continue
            if name not in self._ixnet_infos:
                continue
            # Same name may present within different object structure
            old_keys = sorted(self._ixnet_infos[name].ixnobject)
            keys = sorted(ixnobject)
            if old_keys != keys:
                continue
            set_names.append(name)
            self._ixnet_infos[name] = IxNetInfo(
                ixnobject,
                self.get_working_dg(names[0]),
                index=index,
                multiplier=names.count(name),
                names=names,
            )


class IxNetInfo(object):
    # index start with 0 and use multiplier for count
    def __init__(
        self, ixnobject, working_dg, index=0, multiplier=1, names=None
    ):
        self.ixnobject = ixnobject
        self.working_dg = working_dg
        self.index = int(index)
        self.multiplier = int(multiplier)
        self.names = [] if names is None else names

    @property
    def xpath(self):
        return self.ixnobject.get("xpath")

    @property
    def href(self):
        return self.ixnobject.get("href")



import pytest


@pytest.mark.skip("Already bgpv4 and bgpv6 testcases are available")
def test_devices(api, utils):
    """This is a BGPv4 demo test script with router ranges"""
    config = api.config()

    tx, rx = config.ports.port(
        name="tx", location=utils.settings.ports[0]
    ).port(name="rx", location=utils.settings.ports[1])

    config.options.port_options.location_preemption = True
    ly = config.layer1.layer1()[-1]
    ly.name = "ly"
    ly.port_names = [tx.name, rx.name]
    ly.speed = utils.settings.speed
    ly.media = utils.settings.media

    tx_device, rx_device = config.devices.device(
        name="tx_device", container_name=tx.name
    ).device(name="rx_device", container_name=rx.name)

    # tx_device config
    tx_eth = tx_device.ethernet
    tx_eth.name = "tx_eth"
    tx_ipv4 = tx_eth.ipv4
    tx_ipv4.name = "tx_ipv4"
    tx_ipv4.address.value = "21.1.1.2"
    tx_ipv4.prefix.value = "24"
    tx_ipv4.gateway.value = "21.1.1.1"
    tx_bgpv4 = tx_ipv4.bgpv4
    tx_bgpv4.name = "tx_bgpv4"
    tx_bgpv4.as_type = "ibgp"
    tx_bgpv4.dut_ipv4_address.value = "22.1.1.1"
    tx_bgpv4.as_number.value = "65200"

    tx_rr = tx_bgpv4.bgpv4_route_ranges.bgpv4routerange()[-1]
    tx_rr.name = "tx_rr"
    tx_rr.address_count = "2000"
    tx_rr.address.value = "200.1.0.1"
    tx_rr.prefix.value = "32"

    tx_v6rr = tx_bgpv4.bgpv6_route_ranges.bgpv6routerange()[-1]
    tx_v6rr.name = "tx_v6rr"
    tx_v6rr.address_count = "1000"
    tx_v6rr.address.value = "10::1"
    tx_v6rr.prefix.value = "64"

    # rx_device config
    rx_eth = rx_device.ethernet
    rx_eth.name = "rx_eth"
    rx_ipv4 = rx_eth.ipv4
    rx_ipv4.name = "rx_ipv4"
    rx_ipv4.address.value = "22.1.1.1"
    rx_ipv4.prefix.value = "24"
    rx_ipv4.gateway.value = "22.1.1.2"
    rx_bgpv4 = rx_ipv4.bgpv4
    rx_bgpv4.name = "rx_bgp"
    rx_bgpv4.as_type = "ibgp"
    rx_bgpv4.dut_ipv4_address.value = "22.1.1.2"
    rx_bgpv4.as_number.value = "65200"

    rx_rr = rx_bgpv4.bgpv4_route_ranges.bgpv4routerange()[-1]
    rx_rr.name = "rx_rr"
    rx_rr.address_count = "1000"
    rx_rr.address.value = "200.2.0.1"
    rx_rr.prefix.value = "32"

    # flow config
    flow = config.flows.flow(name="convergence_test")[-1]
    flow.tx_rx.device.tx_names = [tx_rr.name]
    flow.tx_rx.device.rx_names = [rx_rr.name]

    api.set_config(config)


if __name__ == "__main__":
    pytest.main(["-s", __file__])



def test_pretest(api, utils):
    """
    This test does 2 things
    1) Restore ports to factory defaults and keep them ready for tests
    2) Remove any previous stale sessions in API server

    """
    config = api.config()

    for i in range(len(utils.settings.ports)):
        config.ports.port(name="port%d" % i, location=utils.settings.ports[i])[
            -1
        ]

    config.options.port_options.location_preemption = True
    api.set_config(config)

    # ports cleanup
    api._ixnetwork.Vport.find().ResetPortCpuAndFactoryDefault()

    # remove stale sessions
    for session in api.assistant.TestPlatform.Sessions.find():
        if session.Id != api.assistant.Session.Id:
            session.remove()



# import yaml
# import json
#
#
# class SerDesObj(object):
#     """Comment removed"""
#
#     def __init__(self, data):
#         for name, value in data.items():
#             setattr(self, name, self._wrap(value))
#
#     def _wrap(self, value):
#         if isinstance(value, (tuple, list, set, frozenset)):
#             return type(value)([self._wrap(v) for v in value])
#         else:
#             return SerDesObj(value) if isinstance(value, dict) else value
#
#
# class SerDes(object):
#     def __init__(self):
#         pass
#
#     @staticmethod
#     def to_object(contents):
#         """Given a yaml or json string returns a python object
#         If contents is anything else it will just be returned.
#         """
#         if isinstance(contents, (bytes, str)):
#             return SerDes.to_object(yaml.safe_load(contents))
#         elif isinstance(contents, dict):
#             return SerDesObj(contents)
#         else:
#             return contents
#
#     @staticmethod
#     def to_json(obj):
#         """Given a python object returns a json string"""
#         return json.dumps(obj, indent=2, default=lambda x: x.__dict__)
#
#     @staticmethod
#     def to_yaml(obj):
#         """Given a python object returns a yaml string"""
#         return yaml.dump(obj)
#
#
# if __name__ == "__main__":
#     # start with yaml
#     y = """
# config:
#   description: |
#     A literal description.
#     Type-4bits
#      - {some brace}
#      - [some bracket]
#   ports:
#     - name: port 1
#       location: 1.1.1.1/1
#     - name: port 2
#       location: 1.1.1.1/2
#   flows: []
#   layer1: []
#   captures: []
#     """
#     # get object from yaml
#     o = SerDes.to_object(y)
#     # get json from object
#     j = SerDes.to_json(o)
#     print(j)
#     # get object from json
#     o = SerDes.to_object(j)
#     # get yaml from object
#     y = SerDes.to_yaml(o)
#     print(y)
